import gc
import logging
import pathlib
import timeit
from abc import ABC, abstractmethod
from datetime import datetime
from functools import partial
from typing import List, Optional, Sequence, Tuple, TypeVar, Dict

from joblib import Parallel, delayed

from golem.core.adapter import BaseOptimizationAdapter
from golem.core.dag.graph import Graph
from golem.core.log import default_log, Log
from golem.core.optimisers.fitness import Fitness
from golem.core.optimisers.genetic.operators.operator import EvaluationOperator, PopulationT
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.objective import GraphFunction, ObjectiveFunction
from golem.core.optimisers.opt_history_objects.individual import GraphEvalResult
from golem.core.optimisers.timer import Timer, get_forever_timer
from golem.utilities.serializable import Serializable
from golem.utilities.memory import MemoryAnalytics
from golem.utilities.utilities import determine_n_jobs

# the percentage of successful evaluations,
# at which evolution is not threatened with stagnation at the moment
STAGNATION_EVALUATION_PERCENTAGE = 0.5

EvalResultsList = List[GraphEvalResult]
G = TypeVar('G', bound=Serializable)


class DelegateEvaluator:
    """Interface for delegate evaluator of graphs."""

    @property
    @abstractmethod
    def is_enabled(self) -> bool:
        return False

    @abstractmethod
    def compute_graphs(self, graphs: Sequence[G]) -> Sequence[G]:
        raise NotImplementedError()


class ObjectiveEvaluationDispatcher(ABC):
    """Builder for evaluation operator.
    Takes objective function and decides how to evaluate it over population:
    - defines implementation-specific evaluation policy (e.g. sequential, parallel, async);
    - saves additional metadata (e.g. computation time, intermediate metrics values).
    """

    @abstractmethod
    def dispatch(self, objective: ObjectiveFunction, timer: Optional[Timer] = None) -> EvaluationOperator:
        """Return mapped objective function for evaluating population.

        Args:
            objective: objective function that accepts single individual
            timer: optional timer for stopping the evaluation process

        Returns:
            EvaluationOperator: objective function that accepts whole population
        """
        raise NotImplementedError()

    def set_graph_evaluation_callback(self, callback: Optional[GraphFunction]):
        """Set or reset (with None) post-evaluation callback
        that's called on each graph after its evaluation.

        Args:
            callback: callback to be called on each evaluated graph
        """
        pass

    @staticmethod
    def split_individuals_to_evaluate(individuals: PopulationT) -> Tuple[PopulationT, PopulationT]:
        """Split individuals sequence to evaluated and skipped ones."""
        individuals_to_evaluate = []
        individuals_to_skip = []
        for ind in individuals:
            if ind.fitness.valid:
                individuals_to_skip.append(ind)
            else:
                individuals_to_evaluate.append(ind)
        return individuals_to_evaluate, individuals_to_skip

    @staticmethod
    def apply_evaluation_results(individuals: PopulationT,
                                 evaluation_results: EvalResultsList) -> PopulationT:
        """Applies results of evaluation to the evaluated population.
        Excludes individuals that weren't evaluated."""
        evaluation_results = {res.uid_of_individual: res for res in evaluation_results if res}
        individuals_evaluated = []
        for ind in individuals:
            eval_res = evaluation_results.get(ind.uid)
            if not eval_res:
                continue
            ind.set_evaluation_result(eval_res)
            individuals_evaluated.append(ind)
        return individuals_evaluated


class BaseGraphEvaluationDispatcher(ObjectiveEvaluationDispatcher):
    """Base class for dispatchers that evaluate objective function on population.

    Usage: call `dispatch(objective_function)` to get evaluation function.

    Args:
        adapter: adapter for graphs
        n_jobs: number of jobs for multiprocessing or 1 for no multiprocessing.
        graph_cleanup_fn: function to call after graph evaluation, primarily for memory cleanup.
        delegate_evaluator: delegate graph fitter (e.g. for remote graph fitting before evaluation)
    """

    def __init__(self,
                 adapter: BaseOptimizationAdapter,
                 n_jobs: int = 1,
                 graph_cleanup_fn: Optional[GraphFunction] = None,
                 delegate_evaluator: Optional[DelegateEvaluator] = None):
        self._adapter = adapter
        self._objective_eval = None
        self._cleanup = graph_cleanup_fn
        self._post_eval_callback = None
        self._delegate_evaluator = delegate_evaluator

        self.timer = None
        self.logger = default_log(self)
        self._n_jobs = n_jobs
        self.evaluation_cache = None
        self._reset_eval_cache()

    def dispatch(self, objective: ObjectiveFunction, timer: Optional[Timer] = None) -> EvaluationOperator:
        """Return handler to this object that hides all details
        and allows only to evaluate population with provided objective."""
        self._objective_eval = objective
        self.timer = timer or get_forever_timer()
        return self.evaluate_population

    def set_graph_evaluation_callback(self, callback: Optional[GraphFunction]):
        self._post_eval_callback = callback

    def population_evaluation_info(self, pop_size: int, evaluated_pop_size: int):
        """ Shows the amount of successfully evaluated individuals and total number of individuals in population.
         If there are more that 50% of successful evaluations than it's more likely
         there is no problem in optimization process. """
        if pop_size == 0 or evaluated_pop_size / pop_size <= STAGNATION_EVALUATION_PERCENTAGE:
            success_rate = evaluated_pop_size / pop_size if pop_size != 0 else 0
            self.logger.warning(f"{evaluated_pop_size} individuals out of {pop_size} in previous population "
                                f"were evaluated successfully. {success_rate}% "
                                f"is a fairly small percentage of successful evaluation.")
        else:
            self.logger.message(f"{evaluated_pop_size} individuals out of {pop_size} in previous population "
                                f"were evaluated successfully.")

    @abstractmethod
    def evaluate_population(self, individuals: PopulationT) -> PopulationT:
        raise NotImplementedError()

    def evaluate_single(self, graph: OptGraph, uid_of_individual: str, with_time_limit: bool = True,
                        cache_key: Optional[str] = None,
                        logs_initializer: Optional[Tuple[int, pathlib.Path]] = None) -> GraphEvalResult:

        graph = self.evaluation_cache.get(cache_key, graph)

        if with_time_limit and self.timer.is_time_limit_reached():
            return None
        if logs_initializer is not None:
            # in case of multiprocessing run
            Log.setup_in_mp(*logs_initializer)

        adapted_evaluate = self._adapter.adapt_func(self._evaluate_graph)
        start_time = timeit.default_timer()
        fitness, graph = adapted_evaluate(graph)
        end_time = timeit.default_timer()
        eval_time_iso = datetime.now().isoformat()

        eval_res = GraphEvalResult(
            uid_of_individual=uid_of_individual, fitness=fitness, graph=graph, metadata={
                'computation_time_in_seconds': end_time - start_time,
                'evaluation_time_iso': eval_time_iso
            }
        )
        return eval_res

    def _evaluate_graph(self, domain_graph: Graph) -> Tuple[Fitness, Graph]:
        fitness = self._objective_eval(domain_graph)

        if self._post_eval_callback:
            self._post_eval_callback(domain_graph)
        if self._cleanup:
            self._cleanup(domain_graph)
        gc.collect()

        return fitness, domain_graph

    def evaluate_with_cache(self, population: PopulationT) -> PopulationT:
        reversed_population = list(reversed(population))
        self._remote_compute_cache(reversed_population)
        evaluated_population = self.evaluate_population(reversed_population)
        self._reset_eval_cache()
        return evaluated_population

    def _reset_eval_cache(self):
        self.evaluation_cache: Dict[str, Graph] = {}

    def _remote_compute_cache(self, population: PopulationT):
        self._reset_eval_cache()
        if self._delegate_evaluator and self._delegate_evaluator.is_enabled:
            self.logger.info('Remote fit used')
            restored_graphs = self._adapter.restore(population)
            computed_graphs = self._delegate_evaluator.compute_graphs(restored_graphs)
            self.evaluation_cache = {ind.uid: graph for ind, graph in zip(population, computed_graphs)}


class MultiprocessingDispatcher(BaseGraphEvaluationDispatcher):
    """Evaluates objective function on population using multiprocessing pool
    and optionally model evaluation cache with RemoteEvaluator.

    Usage: call `dispatch(objective_function)` to get evaluation function.

    Args:
        adapter: adapter for graphs
        n_jobs: number of jobs for multiprocessing or 1 for no multiprocessing.
        graph_cleanup_fn: function to call after graph evaluation, primarily for memory cleanup.
        delegate_evaluator: delegate graph fitter (e.g. for remote graph fitting before evaluation)
    """

    def __init__(self,
                 adapter: BaseOptimizationAdapter,
                 n_jobs: int = 1,
                 graph_cleanup_fn: Optional[GraphFunction] = None,
                 delegate_evaluator: Optional[DelegateEvaluator] = None):

        super().__init__(adapter, n_jobs, graph_cleanup_fn, delegate_evaluator)

    def dispatch(self, objective: ObjectiveFunction, timer: Optional[Timer] = None) -> EvaluationOperator:
        """Return handler to this object that hides all details
        and allows only to evaluate population with provided objective."""
        super().dispatch(objective, timer)
        return self.evaluate_with_cache

    def evaluate_population(self, individuals: PopulationT) -> PopulationT:
        individuals_to_evaluate, individuals_to_skip = self.split_individuals_to_evaluate(individuals)
        # Evaluate individuals without valid fitness in parallel.
        n_jobs = determine_n_jobs(self._n_jobs, self.logger)

        parallel = Parallel(n_jobs=n_jobs, verbose=0, pre_dispatch="2*n_jobs")
        eval_func = partial(self.evaluate_single, logs_initializer=Log().get_parameters())
        evaluation_results = parallel(delayed(eval_func)(ind.graph, ind.uid) for ind in individuals_to_evaluate)
        individuals_evaluated = self.apply_evaluation_results(individuals_to_evaluate, evaluation_results)
        # If there were no successful evals then try once again getting at least one,
        # even if time limit was reached
        successful_evals = individuals_evaluated + individuals_to_skip
        self.population_evaluation_info(evaluated_pop_size=len(successful_evals),
                                        pop_size=len(individuals))
        if not successful_evals:
            for single_ind in individuals:
                evaluation_result = eval_func(single_ind.graph, single_ind.uid, with_time_limit=False)
                successful_evals = self.apply_evaluation_results([single_ind], [evaluation_result])
                if successful_evals:
                    break
        MemoryAnalytics.log(self.logger,
                            additional_info='parallel evaluation of population',
                            logging_level=logging.INFO)
        return successful_evals


class SequentialDispatcher(BaseGraphEvaluationDispatcher):
    """Evaluates objective function on population in sequential way.

        Usage: call `dispatch(objective_function)` to get evaluation function.
    """

    def evaluate_population(self, individuals: PopulationT) -> PopulationT:
        individuals_to_evaluate, individuals_to_skip = self.split_individuals_to_evaluate(individuals)
        evaluation_results = [self.evaluate_single(ind.graph, ind.uid) for ind in individuals_to_evaluate]
        individuals_evaluated = self.apply_evaluation_results(individuals_to_evaluate, evaluation_results)
        evaluated_population = individuals_evaluated + individuals_to_skip
        return evaluated_population



from copy import deepcopy
from random import random
from typing import Callable, Union, Tuple, TYPE_CHECKING, Mapping, Hashable, Optional

import numpy as np

from golem.core.dag.graph import Graph
from golem.core.optimisers.adaptive.mab_agents.contextual_mab_agent import ContextualMultiArmedBanditAgent
from golem.core.optimisers.adaptive.mab_agents.mab_agent import MultiArmedBanditAgent
from golem.core.optimisers.adaptive.mab_agents.neural_contextual_mab_agent import NeuralContextualMultiArmedBanditAgent
from golem.core.optimisers.adaptive.operator_agent import \
    OperatorAgent, RandomAgent, MutationAgentTypeEnum
from golem.core.optimisers.adaptive.experience_buffer import ExperienceBuffer
from golem.core.optimisers.genetic.operators.base_mutations import \
    base_mutations_repo, MutationTypesEnum
from golem.core.optimisers.genetic.operators.operator import PopulationT, Operator
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.optimisers.opt_history_objects.parent_operator import ParentOperator
from golem.core.optimisers.optimization_parameters import GraphRequirements, OptimizationParameters
from golem.core.optimisers.optimizer import GraphGenerationParams, AlgorithmParameters

if TYPE_CHECKING:
    from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters

MutationFunc = Callable[[Graph, GraphRequirements, GraphGenerationParams, AlgorithmParameters], Graph]
MutationIdType = Hashable
MutationRepo = Mapping[MutationIdType, MutationFunc]


class Mutation(Operator):
    def __init__(self,
                 parameters: 'GPAlgorithmParameters',
                 requirements: GraphRequirements,
                 graph_gen_params: GraphGenerationParams,
                 mutations_repo: Optional[MutationRepo] = None,
                 ):
        super().__init__(parameters, requirements)
        self.graph_generation_params = graph_gen_params
        self.parameters = parameters
        self._mutations_repo = mutations_repo or base_mutations_repo
        self._operator_agent = self._init_operator_agent(graph_gen_params, parameters, requirements)
        self.agent_experience = ExperienceBuffer(window_size=parameters.window_size)

    @staticmethod
    def _init_operator_agent(graph_gen_params: GraphGenerationParams,
                             parameters: 'GPAlgorithmParameters',
                             requirements: OptimizationParameters):
        kind = parameters.adaptive_mutation_type
        if kind == MutationAgentTypeEnum.default or kind == MutationAgentTypeEnum.random:
            agent = RandomAgent(actions=parameters.mutation_types)
        elif kind == MutationAgentTypeEnum.bandit:
            agent = MultiArmedBanditAgent(actions=parameters.mutation_types,
                                          n_jobs=requirements.n_jobs,
                                          path_to_save=requirements.agent_dir,
                                          decaying_factor=parameters.decaying_factor)
        elif kind == MutationAgentTypeEnum.contextual_bandit:
            agent = ContextualMultiArmedBanditAgent(
                actions=parameters.mutation_types,
                context_agent_type=parameters.context_agent_type,
                available_operations=graph_gen_params.node_factory.get_all_available_operations(),
                n_jobs=requirements.n_jobs,
                decaying_factor=parameters.decaying_factor)
        elif kind == MutationAgentTypeEnum.neural_bandit:
            agent = NeuralContextualMultiArmedBanditAgent(
                actions=parameters.mutation_types,
                context_agent_type=parameters.context_agent_type,
                available_operations=graph_gen_params.node_factory.get_all_available_operations(),
                n_jobs=requirements.n_jobs)
        # if agent was specified pretrained (with instance)
        elif isinstance(parameters.adaptive_mutation_type, OperatorAgent):
            agent = kind
        else:
            raise TypeError(f'Unknown parameter {kind}')
        return agent

    @property
    def agent(self) -> OperatorAgent:
        return self._operator_agent

    def __call__(self, population: Union[Individual, PopulationT]) -> Union[Individual, PopulationT]:
        if isinstance(population, Individual):
            population = [population]

        final_population, mutations_applied, application_attempts = tuple(zip(*map(self._mutation, population)))

        # drop individuals to which mutations could not be applied
        final_population = [ind for ind, init_ind, attempt in zip(final_population, population, application_attempts)
                            if not attempt or ind.graph != init_ind.graph]

        if len(population) == 1:
            return final_population[0] if final_population else final_population

        return final_population

    def _mutation(self, individual: Individual) -> Tuple[Individual, Optional[MutationIdType], bool]:
        """ Function applies mutation operator to graph """
        application_attempt = False
        mutation_applied = None
        for _ in range(self.parameters.max_num_of_operator_attempts):
            new_graph = deepcopy(individual.graph)

            new_graph, mutation_applied = self._apply_mutations(new_graph)
            if mutation_applied is None:
                continue
            application_attempt = True
            is_correct_graph = self.graph_generation_params.verifier(new_graph)
            if is_correct_graph:
                parent_operator = ParentOperator(type_='mutation',
                                                 operators=mutation_applied,
                                                 parent_individuals=individual)
                individual = Individual(new_graph, parent_operator,
                                        metadata=self.requirements.static_individual_metadata)
                break
            else:
                # Collect invalid actions
                self.agent_experience.collect_experience(individual, mutation_applied, reward=-1.0)
        else:
            self.log.debug('Number of mutation attempts exceeded. '
                           'Please check optimization parameters for correctness.')
        return individual, mutation_applied, application_attempt

    def _sample_num_of_mutations(self) -> int:
        # most of the time returns 1 or rarely several mutations
        if self.parameters.variable_mutation_num:
            num_mut = max(int(round(np.random.lognormal(0, sigma=0.5))), 1)
        else:
            num_mut = 1
        return num_mut

    def _apply_mutations(self, new_graph: Graph) -> Tuple[Graph, Optional[MutationIdType]]:
        """Apply mutation 1 or few times iteratively"""
        mutation_type = self._operator_agent.choose_action(new_graph)
        mutation_applied = None
        for _ in range(self._sample_num_of_mutations()):
            new_graph, applied = self._adapt_and_apply_mutation(new_graph, mutation_type)
            if applied:
                mutation_applied = mutation_type
                is_custom_mutation = isinstance(mutation_type, Callable)
                if is_custom_mutation:  # custom mutation occurs once
                    break
        return new_graph, mutation_applied

    def _adapt_and_apply_mutation(self, new_graph: Graph, mutation_type) -> Tuple[Graph, bool]:
        applied = self._will_mutation_be_applied(mutation_type)
        if applied:
            # get the mutation function and adapt it
            mutation_func = self._get_mutation_func(mutation_type)
            new_graph = mutation_func(new_graph, requirements=self.requirements,
                                      graph_gen_params=self.graph_generation_params,
                                      parameters=self.parameters)
        return new_graph, applied

    def _will_mutation_be_applied(self, mutation_type: Union[MutationTypesEnum, Callable]) -> bool:
        return random() <= self.parameters.mutation_prob and mutation_type is not MutationTypesEnum.none

    def _get_mutation_func(self, mutation_type: Union[MutationTypesEnum, Callable]) -> Callable:
        if isinstance(mutation_type, Callable):
            mutation_func = mutation_type
        else:
            mutation_func = self._mutations_repo[mutation_type]
        adapted_mutation_func = self.graph_generation_params.adapter.adapt_func(mutation_func)
        return adapted_mutation_func



import itertools
import os
from copy import deepcopy
from datetime import datetime
from glob import glob
from os import remove
from typing import Any, List, Sequence, Tuple, Optional

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from golem.core.dag.graph import Graph
from golem.core.log import default_log
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.optimisers.opt_history_objects.opt_history import OptHistory
from golem.core.paths import default_data_dir
from golem.visualisation.graph_viz import GraphVisualizer
from PIL import Image
from imageio import get_writer, v2


class OptHistoryExtraVisualizer:
    """ Implements legacy history visualizations that are not available via `history.show()`
    Args:
        history: history of optimisation
        folder: path to folder to save results of visualization
    """

    def __init__(self, history: OptHistory, folder: Optional[str] = None):
        data_dir = folder or default_data_dir()

        self.save_path = os.path.join(data_dir, 'composing_history')
        if 'composing_history' not in os.listdir(data_dir):
            os.mkdir(self.save_path)
        self.history = history
        self.log = default_log(self)
        self.graphs_imgs = []
        self.convergence_imgs = []
        self.best_graphs_imgs = []
        self.merged_imgs = []
        self.graph_visualizer = GraphVisualizer

    def pareto_gif_create(self,
                          objectives_numbers: Tuple[int, int] = (0, 1),
                          objectives_names: Tuple[str] = ('ROC-AUC', 'Complexity')):
        files = []
        pareto_fronts = self.history.archive_history
        individuals = self.history.generations
        array_for_analysis = individuals if individuals else pareto_fronts
        all_objectives = extract_objectives(array_for_analysis, objectives_numbers)
        min_x, max_x = min(all_objectives[0]) - 0.01, max(all_objectives[0]) + 0.01
        min_y, max_y = min(all_objectives[1]) - 0.01, max(all_objectives[1]) + 0.01
        folder = f'{self.save_path}'
        for i, front in enumerate(pareto_fronts):
            file_name = f'pareto{i}.png'
            visualise_pareto(front, file_name=file_name, save=True, show=False,
                             folder=folder, generation_num=i, individuals=individuals[i],
                             minmax_x=[min_x, max_x], minmax_y=[min_y, max_y],
                             objectives_numbers=objectives_numbers,
                             objectives_names=objectives_names)
            files.append(f'{folder}/{file_name}')

        create_gif_using_images(gif_path=f'{folder}/pareto_history.gif', files=files)
        for file in files:
            os.remove(file)

    def _visualise_graphs(self, graphs: List[Graph], fitnesses: List[float]):
        fitnesses = deepcopy(fitnesses)
        last_best_graph = graphs[0]
        prev_fit = fitnesses[0]
        fig = plt.figure(figsize=(10, 10))
        for ch_id, graph in enumerate(graphs):
            self.graph_visualizer(graph).draw_nx_dag()
            fig.canvas.draw()
            img = figure_to_array(fig)
            self.graphs_imgs.append(img)
            plt.clf()
            if fitnesses[ch_id] > prev_fit:
                fitnesses[ch_id] = prev_fit
            else:
                last_best_graph = graph
            prev_fit = fitnesses[ch_id]
            plt.clf()
            self.graph_visualizer(last_best_graph).draw_nx_dag()
            fig.canvas.draw()
            img = figure_to_array(fig)
            self.best_graphs_imgs.append(img)
            plt.clf()
        plt.close('all')

    def _visualise_convergence(self, fitness_history):
        fitness_history = deepcopy(fitness_history)
        prev_fit = fitness_history[0]
        for fit_id, fit in enumerate(fitness_history):
            if fit > prev_fit:
                fitness_history[fit_id] = prev_fit
            prev_fit = fitness_history[fit_id]
        ts_set = list(range(len(fitness_history)))
        df = pd.DataFrame(
            {'ts': ts_set, 'fitness': [-f for f in fitness_history]})

        fig = plt.figure(figsize=(10, 10))
        plt.rcParams['axes.titlesize'] = 20
        plt.rcParams['axes.labelsize'] = 20
        for ts in ts_set:
            plt.plot(df['ts'], df['fitness'], label='Optimizer')
            plt.xlabel('Evaluation', fontsize=18)
            plt.ylabel('Best metric', fontsize=18)
            plt.axvline(x=ts, color='black')
            plt.legend(loc='upper left')
            fig.canvas.draw()
            img = figure_to_array(fig)
            self.convergence_imgs.append(img)
            plt.clf()
        plt.close('all')

    def visualise_history(self, metric_index: int = 0):
        try:
            self._clean(with_gif=True)
            all_historical_fitness = self.history.all_historical_quality(metric_index)
            historical_graphs = [ind.graph
                                 for ind in list(itertools.chain(*self.history.generations))]
            self._visualise_graphs(historical_graphs, all_historical_fitness)
            self._visualise_convergence(all_historical_fitness)
            self._merge_images()
            self._combine_gifs()
            self._clean()
        except Exception as ex:
            self.log.error(f'Visualisation failed with {ex}')

    def _merge_images(self):
        for i in range(1, len(self.graphs_imgs)):
            im1 = self.graphs_imgs[i]
            im2 = self.best_graphs_imgs[i]
            im3 = self.convergence_imgs[i]
            imgs = (im1, im2, im3)
            merged = np.concatenate(imgs, axis=1)
            self.merged_imgs.append(Image.fromarray(np.uint8(merged)))

    def _combine_gifs(self):
        date_time = datetime.now().strftime('%B-%d-%Y_%H-%M-%S_%p')
        save_path = os.path.join(self.save_path, f'history_visualisation_{date_time}.gif')
        imgs = self.merged_imgs[1:]
        self.merged_imgs[0].save(save_path, save_all=True, append_images=imgs,
                                 optimize=False, duration=0.5, loop=0)
        self.log.info(f"Visualizations were saved to {save_path}")

    def _clean(self, with_gif=False):
        files = glob(f'{self.save_path}*.png')
        if with_gif:
            files += glob(f'{self.save_path}*.gif')
        for file in files:
            remove(file)

    def _create_boxplot(self, individuals: List[Any], generation_num: int = None,
                        objectives_names: Tuple[str] = ('ROC-AUC', 'Complexity'), file_name: str = 'obj_boxplots.png',
                        folder: str = None, y_limits: Tuple[float] = None):
        folder = f'{self.save_path}/boxplots' if folder is None else folder
        fig, ax = plt.subplots()
        ax.set_title(f'Generation: {generation_num}', fontsize=15)
        objectives = objectives_lists(individuals)
        df_objectives = pd.DataFrame({objectives_names[i]: objectives[i] for i in range(len(objectives))})
        sns.boxplot(data=df_objectives, palette="Blues")
        if y_limits:
            plt.ylim(y_limits[0], y_limits[1])
        if not os.path.isdir('../../tmp'):
            os.mkdir('../../tmp')
        if not os.path.isdir(f'{folder}'):
            os.mkdir(f'{folder}')
        path = f'{folder}/{file_name}'
        plt.savefig(path, bbox_inches='tight')

    def boxplots_gif_create(self, objectives_names: Tuple[str] = ('ROC-AUC', 'Complexity')):
        individuals = self.history.generations
        objectives = extract_objectives(individuals)
        objectives = list(itertools.chain(*objectives))
        min_y, max_y = min(objectives), max(objectives)
        files = []
        folder = f'{self.save_path}'
        for generation_num, individuals_in_genaration in enumerate(individuals):
            file_name = f'{generation_num}.png'
            self._create_boxplot(individuals_in_genaration, generation_num, objectives_names,
                                 file_name=file_name, folder=folder, y_limits=(min_y, max_y))
            files.append(f'{folder}/{file_name}')
        create_gif_using_images(gif_path=f'{folder}/boxplots_history.gif', files=files)
        for file in files:
            os.remove(file)
        plt.cla()
        plt.clf()
        plt.close('all')


def visualise_pareto(front: Sequence[Individual],
                     objectives_numbers: Tuple[int, int] = (0, 1),
                     objectives_names: Sequence[str] = ('ROC-AUC', 'Complexity'),
                     file_name: str = 'result_pareto.png', show: bool = False, save: bool = True,
                     folder: str = '../../tmp/pareto',
                     generation_num: int = None,
                     individuals: Sequence[Individual] = None,
                     minmax_x: List[float] = None,
                     minmax_y: List[float] = None):
    pareto_obj_first, pareto_obj_second = [], []
    for ind in front:
        fit_first = ind.fitness.values[objectives_numbers[0]]
        pareto_obj_first.append(abs(fit_first))
        fit_second = ind.fitness.values[objectives_numbers[1]]
        pareto_obj_second.append(abs(fit_second))

    fig, ax = plt.subplots()

    if individuals is not None:
        obj_first, obj_second = [], []
        for ind in individuals:
            fit_first = ind.fitness.values[objectives_numbers[0]]
            obj_first.append(abs(fit_first))
            fit_second = ind.fitness.values[objectives_numbers[1]]
            obj_second.append(abs(fit_second))
        ax.scatter(obj_first, obj_second, c='green')

    ax.scatter(pareto_obj_first, pareto_obj_second, c='red')
    plt.plot(pareto_obj_first, pareto_obj_second, color='r')

    if generation_num is not None:
        ax.set_title(f'Pareto frontier, Generation: {generation_num}', fontsize=15)
    else:
        ax.set_title('Pareto frontier', fontsize=15)
    plt.xlabel(objectives_names[0], fontsize=15)
    plt.ylabel(objectives_names[1], fontsize=15)

    if minmax_x is not None:
        plt.xlim(minmax_x[0], minmax_x[1])
    if minmax_y is not None:
        plt.ylim(minmax_y[0], minmax_y[1])
    fig.set_figwidth(8)
    fig.set_figheight(8)
    if save:
        if not os.path.isdir('../../tmp'):
            os.mkdir('../../tmp')
        if not os.path.isdir(f'{folder}'):
            os.mkdir(f'{folder}')

        path = f'{folder}/{file_name}'
        plt.savefig(path, bbox_inches='tight')
    if show:
        plt.show()

    plt.cla()
    plt.clf()
    plt.close('all')


def create_gif_using_images(gif_path: str, files: List[str]):
    with get_writer(gif_path, mode='I', duration=0.5) as writer:
        for filename in files:
            image = v2.imread(filename)
            writer.append_data(image)


def extract_objectives(individuals: List[List[Any]], objectives_numbers: Tuple[int, ...] = None,
                       transform_from_minimization=True):
    if not objectives_numbers:
        objectives_numbers = [i for i in range(len(individuals[0][0].fitness.values))]
    all_inds = list(itertools.chain(*individuals))
    all_objectives = [[ind.fitness.values[i] for ind in all_inds] for i in objectives_numbers]
    if transform_from_minimization:
        transformed_objectives = []
        for obj_values in all_objectives:
            are_objectives_positive = all(np.array(obj_values) > 0)
            if not are_objectives_positive:
                transformed_obj_values = list(np.array(obj_values) * (-1))
            else:
                transformed_obj_values = obj_values
            transformed_objectives.append(transformed_obj_values)
    else:
        transformed_objectives = all_objectives
    return transformed_objectives


def figure_to_array(fig):
    img = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8)
    img = img.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    return img


def objectives_lists(individuals: List[Any], objectives_numbers: Tuple[int] = None):
    num_of_objectives = len(objectives_numbers) if objectives_numbers else len(individuals[0].fitness.values)
    objectives_numbers = objectives_numbers if objectives_numbers else [i for i in range(num_of_objectives)]
    objectives_values_set = [[] for _ in range(num_of_objectives)]
    for obj_num in range(num_of_objectives):
        for individual in individuals:
            value = individual.fitness.values[objectives_numbers[obj_num]]
            objectives_values_set[obj_num].append(value if value > 0 else -value)
    return objectives_values_set



from random import shuffle

from golem.core.optimisers.genetic.operators.operator import PopulationT, Operator
from golem.utilities.data_structures import ComparableEnum as Enum


class ElitismTypesEnum(Enum):
    keep_n_best = 'keep_n_best'
    replace_worst = 'replace_worst'
    none = 'none'


class Elitism(Operator):
    def __call__(self, best_individuals: PopulationT, new_population: PopulationT) -> PopulationT:
        elitism_type = self.parameters.elitism_type
        if elitism_type is ElitismTypesEnum.none or not self._is_elitism_applicable():
            return new_population
        elif elitism_type is ElitismTypesEnum.keep_n_best:
            return self.keep_n_best_elitism(best_individuals, new_population)
        elif elitism_type is ElitismTypesEnum.replace_worst:
            return self.replace_worst_elitism(best_individuals, new_population)
        else:
            raise ValueError(f'Required elitism type not found: {elitism_type}')

    def _is_elitism_applicable(self) -> bool:
        if self.parameters.multi_objective:
            return False
        return self.parameters.pop_size >= self.parameters.min_pop_size_with_elitism

    @staticmethod
    def keep_n_best_elitism(best_individuals: PopulationT, new_population: PopulationT) -> PopulationT:
        final_population = []
        final_population += best_individuals
        new_unique_inds = [ind for ind in new_population if ind not in best_individuals]
        if new_unique_inds:
            shuffle(new_unique_inds)
            remain_n = len(new_population) - len(best_individuals)
            final_population += new_unique_inds[:remain_n]
        return final_population

    @staticmethod
    def replace_worst_elitism(best_individuals: PopulationT, new_population: PopulationT) -> PopulationT:
        population = best_individuals + new_population
        # sort in descending order (Fitness(10) > Fitness(11))
        sorted_ascending_population = sorted(population, key=lambda individual: individual.fitness, reverse=True)
        return sorted_ascending_population[:len(new_population)]
